#!/bin/bash
#############################################################
# Linux Agent Installer Script
#############################################################

# Set umask so that file and directories have max permission of 644 and 755 respectively
umask 022

# Constants
readonly ERROR=0
readonly WARN=1
readonly INFO=2
readonly DEBUG=3
readonly SCRIPT_NAME="${0##*/}"
readonly IS_ROOT=$([[ $(id -u) -eq 0 ]] && echo true || echo false)
readonly EFCTV_USR=$(id -un)
readonly REQ_CMDS="chmod chown cp cpio cut grep gzip hash id ls mkdir mv sed"
readonly OS_NAME=$(grep "^ID=" /etc/os-release | cut -d = -f 2 | tr -d \" | tr '[:upper:]' '[:lower:]')
readonly OS_LIKE=$(grep "^ID_LIKE=" /etc/os-release | cut -d = -f 2 | tr -d \" | tr '[:upper:]' '[:lower:]')
readonly SYS_BIN_DIR="/usr/local/bin"
readonly SYSTEMD_DIR="/usr/lib/systemd/system"
readonly SYSTEMD_UNIT_FILE="crashplan.service"
readonly APP_VERSION_FILE="c42.version.properties"
readonly INSTALL_DFLT_FILE="install.defaults"
readonly INSTALL_VAR_FILE="install.vars"

SCRIPT_DIR="${0:0:${#0}-${#SCRIPT_NAME}}"
PARENT_DIR=
APP_DIR=
BIN_DIR=
MANIFEST_DIR=

# Config
APP_BASENAME=
APP_BASENAME_LOWER=
SVC_SCRIPT=
SVC_SCRPT_PTH=
GUI_SCRIPT=
GUI_SCRPT_PTH=
GUI_DSPLY_PTH=
CLOAKED=false
APP_DEPLOYMENT_ARGS=""
LOG_LEVEL=${INFO}
LOG_FILE=
QUIET="false"
USER_INSTALL="false"
TARGET_USER=
UPGRADE="false"

function log_level_to_string() {
  case $1 in
  ${ERROR})
    echo "Error"
    ;;
  ${WARN})
    echo "Warning"
    ;;
  ${INFO})
    echo "Info"
    ;;
  ${DEBUG})
    echo "Debug"
    ;;
  *)
    echo "Unknown"
    ;;
  esac
}

function log() {
  if [[ $1 -gt ${LOG_LEVEL} ]] || [[ -z "${LOG_FILE}" && "${QUIET}" == "true" ]]; then
    return 0
  fi

  local time_stamp=$(date)
  local level=$(log_level_to_string "$1")
  local message=$2
  local log_message="$time_stamp: $level : $message"

  if [[ -n "${LOG_FILE}" ]]; then
    echo "$log_message" >>"${LOG_FILE}"
  elif [[ "${QUIET}" == "false" ]]; then
    echo "$log_message"
  fi
}

function usage() {
  echo "Usage: $0 [-chqv] [-d \"args ...]\" [-l file] [-u user] [-x path]"
  echo "${APP_BASENAME} Installer"
  echo
  echo "    -c                  Cloaked install."
  echo "    -d [arg1 arg2 ...]  Deployment options to be passed to the service on first launch."
  echo "    -h                  Show this message."
  echo "    -l [log_file]       Relative or absolute path to log file [default: stdout]."
  echo "    -q                  Do not prompt user for input, use defaults."
  echo "    -u [user]           Install for the user."
  echo "    -v                  Add debug lines to the output."
  echo "    -x [path]           Install path."
  echo
}

return_absolute_path() {
  local path="${1}"
  if [[ "${path}" = /* ]]; then
    echo "${path}"
  else
    echo "$(
      cd ${path}
      pwd
    )"
  fi
}

function process_input() {
  while getopts "cd:hl:qu:vx:" opt; do
    case "${opt}" in
    c) CLOAKED=true ;;
    d) APP_DEPLOYMENT_ARGS="$OPTARG" ;;
    h)
      usage
      exit 0
      ;;
    l)
      LOG_FILE="$OPTARG"
      if [[ -f ${LOG_FILE} ]]; then
        log ${INFO} "============================================================================"
        log ${INFO} "                            Starting Install"
        log ${INFO} "============================================================================"
      fi
      ;;
    q) QUIET="true" ;;
    u)
      USER_INSTALL="true"
      TARGET_USER="$OPTARG"
      ;;
    v) LOG_LEVEL=${DEBUG} ;;
    x)
      PARENT_DIR="$(return_absolute_path "${OPTARG%/}")"
      readonly PARENT_DIR
      ;;
    \?)
      usage
      exit 1
      ;;
    esac
  done
  log ${DEBUG} "${SCRIPT_NAME} ran with arguments $*"

  shift $((OPTIND - 1))

  readonly QUIET
  readonly LOG_FILE
  readonly LOG_LEVEL
  readonly SCRIPT_DIR="$(return_absolute_path "${SCRIPT_DIR}")"

  log ${DEBUG} "Script ran from directory: '${SCRIPT_DIR}'"
}

function verify_cmds() {
  local missing_cmds
  for cmd in ${REQ_CMDS}; do
    if ! hash "${cmd}" &>/dev/null; then
      missing_cmds="${missing_cmds} '${cmd}'"
    fi
  done
  if [[ -n "${missing_cmds}" ]]; then
    log ${ERROR} "Missing command(s) required for install:${missing_cmds}"
    exit 1
  fi
}

function get_yn_response() {
  local default="$2"
  if [[ "${QUIET}" == "false" ]]; then
    read -r yn
  fi

  if [[ -z "${yn}" ]]; then
    [[ -n "${default}" ]] && yn="${default}" || yn=n
  fi

  case ${yn} in
  [Yy] | [Yy][Ee][Ss]) yn=y ;;
  *) yn=n ;;
  esac

  log ${DEBUG} "Response is: ${yn}"
  eval "$1=${yn}"
}

# variables defined in INSTALL_DFLT_FILE
# APP_BASENAME
# DIR_BASENAME
function source_install_defaults() {
  if [[ ! -f "${SCRIPT_DIR}/${INSTALL_DFLT_FILE}" ]]; then
    log ${ERROR} "${SCRIPT_DIR}/${INSTALL_DFLT_FILE} missing!"
    exit 1
  fi

  source "${SCRIPT_DIR}/${INSTALL_DFLT_FILE}"

  readonly APP_BASENAME
  readonly DIR_BASENAME

  readonly APP_BASENAME_LOWER="$(echo "${APP_BASENAME}" | tr '[:upper:]' '[:lower:]')"
}

function set_service_scripts() {
  readonly SVC_SCRIPT="service.sh"
  readonly GUI_SCRIPT="desktop.sh"
}

function check_for_user_install() {
  if [[ -n "${PARENT_DIR}" && "${PARENT_DIR}" == /home/* ]]; then
    log ${INFO} "Detected '/home/*' in provided path '${PARENT_DIR}'. Assuming user install."
    TARGET_USER="${EFCTV_USR}"
    USER_INSTALL="true"
  fi
  readonly USER_INSTALL
  readonly TARGET_USER
}

function verify_user() {
  log ${DEBUG} "Installing as '${EFCTV_USR}'"
  log ${DEBUG} "User install: ${USER_INSTALL}"

  if [[ "${USER_INSTALL}" == "true" ]] && ! id -u "${TARGET_USER}" &>/dev/null; then
    log ${ERROR} "User doesn't exist: '${TARGET_USER}'"
    exit 1
  fi

  if [[ "${USER_INSTALL}" == "false" ]] && [[ "${IS_ROOT}" == "false" ]]; then
    log ${ERROR} "Insufficient privileges. Running as '${EFCTV_USR}'"
    log ${ERROR} "Please run as root or specify user only install."
    exit 1
  fi

  [[ "${USER_INSTALL}" == "true" ]] && log ${DEBUG} "Installing for target user: ${TARGET_USER}"
}

function alert_with_error() {
  local err=$1
  if [[ "${QUIET}" == "false" ]]; then
    echo "${err}"
    return 1
  fi

  log ${ERROR} "${err}"
  exit 1
}

function prompt_to_create_dir_if_not_found() {
  local des_dir=$1
  [[ -d "${des_dir}" ]] && return 0

  [[ "${QUIET}" == "false" ]] && echo -n "${des_dir} does not exist. Create $1? (y/n) [y]: "
  get_yn_response YN y

  if [[ "${YN}" == "n" ]]; then
    alert_with_error "Rejected creation of ${des_dir}."
    return 1
  fi

  mkdir -p "${des_dir}" && return 0
  alert_with_error "Failed to create ${des_dir}. Please check your configuration."
}

function request_dir_from_user() {
  local prompt="$2"
  local default_dir="$3"
  local sub_dir="$4"
  local install_dir=
  while true; do
    if [[ "${QUIET}" == "false" ]]; then
      echo
      echo -n "${prompt} [${default_dir}]: "
      read -r install_dir
    fi
    if [[ -n "${install_dir}" ]]; then
      install_dir="$(return_absolute_path "${install_dir}")"
      install_dir="${install_dir%/}"
    else
      install_dir="${default_dir}"
    fi

    [[ -n "${sub_dir}" ]] && install_dir="${install_dir}/${sub_dir}"

    if prompt_to_create_dir_if_not_found "${install_dir}"; then
      break
    fi
  done
  eval "$1=\${install_dir}"
}

function prompt_to_review_install_locations() {
  log ${DEBUG} "Reviewing install directories."
  if [[ "${QUIET}" == "false" ]]; then
    echo ""
    echo "Your selections:"
    echo "${APP_BASENAME} will install to: ${APP_DIR}"
    echo ""
    echo -n "Is this correct? (y/n) [y]: "
  fi

  get_yn_response YN y
  [[ "${YN}" == y ]]
}

function prompt_for_install_locations() {
  while true; do
    request_dir_from_user APP_DIR \
      "What parent directory do you wish to install ${APP_BASENAME} into?" \
      "${PARENT_DIR}" \
      "${DIR_BASENAME}"
    prompt_to_review_install_locations && break
  done
}

function set_install_locations() {
  if [[ "${USER_INSTALL}" == "false" ]]; then
    [[ -z "${PARENT_DIR}" ]] && PARENT_DIR="/usr/local"
  else
    [[ -z "${PARENT_DIR}" ]] && PARENT_DIR=$(grep "${TARGET_USER}" /etc/passwd | cut -d : -f 6)
  fi
  APP_DIR="${PARENT_DIR}/${DIR_BASENAME}"

  log ${DEBUG} "Skipping user input for install directories: ${QUIET}"
  prompt_for_install_locations

  readonly APP_DIR
  readonly BIN_DIR="${APP_DIR}/bin"
  readonly MANIFEST_DIR="${APP_DIR}/manifest"
}

function get_value_of_key_from_file() {
  local key="$2"
  local kv_file="$3"
  if [[ ! -f "${kv_file}" ]]; then
    log ${WARN} "Cannot locate key-value file: '${kv_file}'"
    return 1
  fi
  local value=$(cat "${kv_file}" | grep "${key}=" | cut -f 2 -d '=')
  if [[ -z "${value}" ]]; then
    log ${WARN} "Failed to retrieve value from file: '${kv_file}', key: '${key}'"
    return 1
  fi
  eval "$1=${value}"
}

function is_installed_build_newer() {
  local installed_build
  local upgrade_build
  get_value_of_key_from_file installed_build "build_number" "${APP_DIR}/${APP_VERSION_FILE}"
  get_value_of_key_from_file upgrade_build "build_number" "${SCRIPT_DIR}/${APP_VERSION_FILE}"
  log ${DEBUG} "Installed application build number: ${installed_build}"
  log ${DEBUG} "Upgrade application build number: ${upgrade_build}"
  [[ "${upgrade_build}" -ge "${installed_build}" ]] && return 1 || return 0
}

function is_installed_version_newer() {
  local installed_version
  local upgrade_version
  get_value_of_key_from_file installed_version "cp_version" "${APP_DIR}/${APP_VERSION_FILE}"
  get_value_of_key_from_file upgrade_version "cp_version" "${SCRIPT_DIR}/${APP_VERSION_FILE}"
  log ${DEBUG} "Installed application version: ${installed_version}"
  log ${DEBUG} "Upgrade application version: ${upgrade_version}"
  [[ "${upgrade_version}" -lt "${installed_version}" ]] && return 0
  [[ "${upgrade_version}" -gt "${installed_version}" ]] && return 1
  [[ "${upgrade_version}" -eq "${installed_version}" ]] && is_installed_build_newer
}

function prev_install_exists() {
  [[ -f "${APP_DIR}/install.vars" ]]
}

function usr_would_like_to_upgrade() {
  if [[ "${QUIET}" == false ]]; then
    echo "${APP_BASENAME} appears to already be installed in the specified location:"
    echo "  ${APP_DIR}"
    echo -n "Would you like to upgrade? (y/n) [y]: "
  fi
  get_yn_response YN y
  [[ "${YN}" == y ]]
}

function attempt_to_upgrade() {
  log ${INFO} "Upgrading ${APP_BASENAME}..."
  local user_arg=
  local log_arg=
  local quiet_arg=
  [[ "${USER_INSTALL}" == "true" ]] && user_arg="-u ${TARGET_USER}"
  [[ -n "${LOG_FILE}" ]] && log_arg="-l ${LOG_FILE}"
  [[ "${QUIET}" == "true" ]] && quiet_arg="-q"
  if ! /bin/bash uninstall.sh -v -y -i "${APP_DIR} -x" "${user_arg}" "${log_arg}" "${quiet_arg}"; then
    log ${ERROR} "Uninstall failed"
    exit 1
  fi
  log ${INFO} "Uninstall succeeded. Continuing."
  return 0
}

function check_and_handle_prev_install() {
  if ! prev_install_exists; then
    log ${INFO} "Did not detect an existing installation. Continuing."
    return 0
  fi

  if is_installed_version_newer; then
    log ${INFO} "Installed version is newer. Nothing to be done."
    exit 0
  fi

  usr_would_like_to_upgrade && attempt_to_upgrade && return 0

  log ${INFO} "Rejecting upgrade. Exiting install script."
  exit 1
}

function prompt_install_pkg_if_not_present() {
  local pkg="$1"
  local search_command=$2
  local install_command=$3
  log ${DEBUG} "Checking for required dependency: '${pkg}'"
  if ${search_command} &>/dev/null; then
    log ${INFO} "${pkg} dependency found."
    return 0
  fi
  if [[ "${QUIET}" == "false" ]]; then
    echo "${pkg} is required by ${APP_BASENAME} and is not installed."
    echo "Would you like to install it? (y/n) [y]: "
  fi
  get_yn_response YN y
  if [[ ${YN} == "y" ]]; then
    if ${install_command} &>/dev/null; then
      log ${INFO} "${pkg} was installed."
      return 0
    else
      log ${WARN} "${pkg} failed to install."
    fi
  fi
  log ${ERROR} "Unable to install, missing required dependency ${pkg}."
  exit 1
}

function apt_check_for_pkg() {
  local pkg="$1"
  local search_command="dpkg -s ${pkg}"
  local install_command="apt install ${pkg} -y"
  prompt_install_pkg_if_not_present "${pkg}" "${search_command}" "${install_command}"
}

function yum_check_for_pkg() {
  local pkg="$1"
  local search_command="yum list installed ${pkg}"
  local install_command="yum install ${pkg} -y"
  prompt_install_pkg_if_not_present "${pkg}" "${search_command}" "${install_command}"
}

function install_app_prereqs() {
  # RHEL has an issue where the check can freeze on upgrade (AGT-2087). The checks would have
  # been done on install and we would be in a broken state should the below pkgs have been
  # removed after, so we can skip for upgrades.
  if [[ "${UPGRADE}" == true ]]; then
    log ${INFO} "Skipping install requirements check, due to upgrading."
    return 0
  fi

  log ${INFO} "Checking prerequisites for ${OS_NAME}..."

  [[ "${OS_NAME}" == "rhel" ]] || [[ "${OS_LIKE}" == *rhel* ]] && yum_check_for_pkg libXScrnSaver
  [[ "${OS_NAME}" =~ ^ubuntu$|^debian$ ]] || [[ "${OS_LIKE}" == *debian* ]] && apt_check_for_pkg libxss1
}

function create_dir_if_not_found() {
  local des_dir="$1"
  [[ -d "${des_dir}" ]] && return 0

  if ! mkdir -p "${des_dir}"; then
    log ${ERROR} "Failed to create ${des_dir}. Please check your configuration."
    exit 1
  fi
  log ${DEBUG} "Created directory '${des_dir}'"
}

function create_install_dirs() {
  create_dir_if_not_found "${APP_DIR}"
  [[ -n "${BIN_DIR}" ]] && create_dir_if_not_found "${BIN_DIR}"
  create_dir_if_not_found "${MANIFEST_DIR}"
}

function log_install_location_info() {
  log ${INFO} ""
  log ${INFO} "${APP_BASENAME} install info:"
  log ${INFO} "  Application directory: '${APP_DIR}'"
  [[ -n "${BIN_DIR}" ]] &&
    log ${INFO} "  Binary directory: '${BIN_DIR}'"
  log ${INFO} "  Manifest directory: '${MANIFEST_DIR}'"
  [[ "${USER_INSTALL}" == "false" ]] &&
    log ${INFO} "  Systemd unit file directory: ${SYSTEMD_DIR}"
  log ${INFO} ""
}

function write_install_vars() {
  cat <<EOF >"${install_var_file}"
TARGETDIR=${APP_DIR}
BINSDIR=${BIN_DIR}
MANIFESTDIR=${MANIFEST_DIR}
INSTALLDATE=${now}
JAVACOMMON=${java_common}
EOF
}

function create_install_vars() {
  local now=$(date +%Y%m%d)
  local java_common="${APP_DIR}/jre/bin/java"
  local install_var_file="${APP_DIR}/${INSTALL_VAR_FILE}"
  write_install_vars
  [[ "${USER_INSTALL}" == "false" ]] &&
    echo "SYSTEMD_DIR=${SYSTEMD_DIR}" >>"${install_var_file}"
  cat "${SCRIPT_DIR}/${INSTALL_DFLT_FILE}" >>"${install_var_file}"
}

function unpack_archive() {
  local app_archive=$(ls "${SCRIPT_DIR}"/*.cpi 2>/dev/null)
  if [[ ! -f "${app_archive}" ]]; then
    log ${ERROR} "Failed to locate app archive directory in ${SCRIPT_DIR}"
    exit 1
  fi

  log ${INFO} "Unpacking ${app_archive} ..."
  if ! $(
    cd "${APP_DIR}"
    gzip -dc "${app_archive}" | cpio -i --no-preserve-owner &>/dev/null
  ); then
    log ${ERROR} "Failed to extract archived file: ' $(pwd)/${app_archive}'"
    exit 1
  fi
  find "${APP_DIR}" -type d -exec chmod 755 "{}" \; &>/dev/null
  find "${APP_DIR}" -type f \
    ! -name libjsig.so ! -name restore-tool -exec chmod 744 "{}" \; &>/dev/null
}

function setup_electron() {
  mv -v "${APP_DIR}"/*.asar "${APP_DIR}/electron/resources" &>/dev/null
  chmod 755 "${APP_DIR}/electron/${APP_BASENAME_LOWER}"
  chmod 4755 "${APP_DIR}/electron/chrome-sandbox"
}

function update_manifest_path() {
  #update the configs for file storage
  if grep "<manifestPath>.*</manifestPath>" "${APP_DIR}/conf/default.service.xml" >/dev/null; then
    sed -i "s|<manifestPath>.*</manifestPath>|<manifestPath>${MANIFEST_DIR}</manifestPath>|g" "${APP_DIR}/conf/default.service.xml"
  else
    sed -i "s|<backupConfig>|<backupConfig>\n\t\t\t<manifestPath>${MANIFEST_DIR}</manifestPath>|g" "${APP_DIR}/conf/default.service.xml"
  fi
}

function change_log_dir_perms() {
  chmod 755 "${APP_DIR}/log"
}

function install_gui_script() {
  # desktop init script
  GUI_SCRPT_PTH="${BIN_DIR}/${GUI_SCRIPT}"

  if [[ ! -f "${SCRIPT_DIR}/scripts/${GUI_SCRIPT}" ]]; then
    log ${WARN} "Desktop script doesn't exist: '${SCRIPT_DIR}/scripts/${GUI_SCRIPT}'"
    return 1
  fi

  log ${DEBUG} "Copying '${SCRIPT_DIR}/scripts/${GUI_SCRIPT}' to '${GUI_SCRPT_PTH}'"
  cp "${SCRIPT_DIR}/scripts/${GUI_SCRIPT}" "${GUI_SCRPT_PTH}"
  chmod 755 "${GUI_SCRPT_PTH}"

  if [[ "${USER_INSTALL}" == "false" ]]; then
    ln -s "${GUI_SCRPT_PTH}" "${SYS_BIN_DIR}/${APP_BASENAME_LOWER}" 2>/dev/null &&
      log ${INFO} "Created symbolic link from '${GUI_SCRPT_PTH}' to '${SYS_BIN_DIR}/${APP_BASENAME_LOWER}'" ||
      log ${WARN} "Failed to create symbolic link from '${GUI_SCRPT_PTH}' to '${SYS_BIN_DIR}/${APP_BASENAME_LOWER}'"
  fi

  readonly GUI_DSPLY_PTH="$([[ "${USER_INSTALL}" == "false" ]] &&
    echo "${SYS_BIN_DIR}/${APP_BASENAME_LOWER}" ||
    echo "${GUI_SCRPT_PTH}")"
}

function install_service_script() {
  SVC_SCRPT_PTH="${BIN_DIR}/${SVC_SCRIPT}"
  local svc_script_pkg_pth="${SCRIPT_DIR}/scripts/${SVC_SCRIPT}"
  if [[ ! -f "${svc_script_pkg_pth}" ]]; then
    log ${WARN} "Service script doesn't exist: '${svc_script_pkg_pth}'"
    return 1
  fi

  log ${DEBUG} "Copying '${svc_script_pkg_pth}' to '${SVC_SCRPT_PTH}'"
  cp "${svc_script_pkg_pth}" "${SVC_SCRPT_PTH}"
  chmod 744 "${SVC_SCRPT_PTH}"
}

function write_systemd_script() {
  [ ! -d ${SYSTEMD_DIR} ] && mkdir -p "${SYSTEMD_DIR}" 2>/dev/null
  cat <<EOF >"${SYSTEMD_DIR}/${SYSTEMD_UNIT_FILE}"
[Unit]
Description=${APP_BASENAME} Service
After=network-online.target
Wants=network-online.target

[Service]
Type=forking
Restart=no
TimeoutSec=5min
IgnoreSIGPIPE=no
KillMode=process
GuessMainPID=no
RemainAfterExit=yes
SuccessExitStatus=5 6
ExecStart="${SVC_SCRPT_PTH}" start
ExecStop="${SVC_SCRPT_PTH}" stop
ExecReload="${SVC_SCRPT_PTH}" reload

[Install]
WantedBy=multi-user.target
EOF
}

function install_systemd_script() {
  [[ "${USER_INSTALL}" == "true" ]] && return 0
  write_systemd_script
  chmod 644 "${SYSTEMD_DIR}/${SYSTEMD_UNIT_FILE}"
  systemctl daemon-reload &>/dev/null
  systemctl enable "${APP_BASENAME_LOWER}.service" &>/dev/null
}

##
# Gets inotify watch file limit
#
# Arguments:
# $1: integer | Out param inidicating the intotify watch limit
##
function get_inotify_limit() {
  local inotify_limit_file="/proc/sys/fs/inotify/max_user_watches"
  local inotify_watches
  if [[ -f "$inotify_limit_file" ]]; then
    inotify_watches=$(cat "$inotify_limit_file")
  fi

  if [[ -z "${inotify_watches}" ]]; then
    return 1
  fi

  eval "$1=\$inotify_watches"
}

function set_inotify_limit() {
  local inotify_watches=$1
  local sysctl_file="/etc/sysctl.conf"
  if ! [[ -f "$sysctl_file" ]]; then
    log ${WARN} "'$sysctl_file' does not exist."
    return 1
  fi

  if grep fs.inotify.max_user_watches /etc/sysctl.conf | grep -v "#"; then
    sed -e '/fs.inotify.max_user_watches=/ s/^#*/#/' -i /etc/sysctl.conf
  fi
  echo "fs.inotify.max_user_watches=$inotify_watches" >>/etc/sysctl.conf
  if ! sysctl -p -q &>/dev/null; then
    log ${WARN} "Failed to load in sysctl settings from '${sysctl_file}'"
    log ${WARN} "Please load sysctl setting for modified inotify max user watches limit to take effect."
  fi
}

function configure_inotify_settings() {
  local rec_inotify_watches=1048576
  local sys_inotify_watches

  if ! get_inotify_limit sys_inotify_watches; then
    log ${WARN} "Could not determine inotify limit."
    log ${WARN} "${APP_BASENAME} recommends setting the inotify limit to ${rec_inotify_watches}."
    return 1
  fi

  if [[ "${sys_inotify_watches}" -ge "${rec_inotify_watches}" ]]; then
    return 0
  fi

  if [[ "${QUIET}" == false ]]; then
    echo -e "\nFor an optimal realtime file watcher experience, "
    echo "${APP_BASENAME} recommends setting the inotify limit to ${rec_inotify_watches}."
    echo "Your current limit is ${sys_inotify_watches}."
    echo -n "Set inotify limit to ${rec_inotify_watches}? (y/n) [y] "
  fi

  get_yn_response YN y
  if [[ "${YN}" == "y" ]]; then
    if ! set_inotify_limit "$rec_inotify_watches"; then
      log ${WARN} "Could not set inotify limit."
      log ${WARN} "${APP_BASENAME} recommends setting the inotify limit to $rec_inotify_watches"
      return 1
    else
      log ${INFO} "Successfully set inotify limit."
    fi
  fi
}

function setup_app_launcher() {
  local app_launcher="$1"
  local desktop_icon="${BIN_DIR}/icon_app3.png"
  local desktop_file="$(ls "${SCRIPT_DIR}"/scripts/*.desktop 2>/dev/null)"

  if [[ ! -f "${desktop_icon}" ]]; then
    log ${WARN} "Desktop icon does not exist: '${desktop_icon}'"
    return 1
  fi
  if [[ ! -f "${desktop_file}" ]]; then
    log ${WARN} "Desktop file does not exist: '${desktop_file}'"
    return 1
  fi

  cp "${desktop_file}" "${app_launcher}"
  sed -i "s|Exec=|Exec=${GUI_SCRPT_PTH}|" "${app_launcher}"
  sed -i "s|Icon=.*|Icon=${desktop_icon}|" "${app_launcher}"
}

function install_launcher() {
  local launcher_file_name="${APP_BASENAME_LOWER}.desktop"
  local app_launcher="$([[ "${USER_INSTALL}" == "false" ]] &&
    echo "/usr/share/applications/${launcher_file_name}" ||
    echo "${PARENT_DIR}/.local/share/applications/${launcher_file_name}")"
  setup_app_launcher "${app_launcher}"

  if ! configure_inotify_settings; then
    log ${WARN} "Configuration of inotify watch limit failed."
  fi
}

function log_resource_location() {
  log ${INFO} ""
  log ${INFO} "Important directories:"
  log ${INFO} "  Installation:"
  log ${INFO} "    ${APP_DIR}"
  log ${INFO} "  Logs:"
  log ${INFO} "    ${APP_DIR}/log"
  log ${INFO} "    ~/.${APP_BASENAME_LOWER}/log"

  log ${INFO} ""
  log ${INFO} "Service Start Script:"
  log ${INFO} "  ${SVC_SCRPT_PTH} start|stop"
  log ${INFO} ""
  log ${INFO} "To start the Desktop UI:"
  log ${INFO} "  ${GUI_DSPLY_PTH}"
}

function start_service() {
  local success=false
  readonly user_install_arg="$([[ "${USER_INSTALL}" == "false" ]] &&
    echo "-Duser.install=0" ||
    echo "-Duser.install=1")"
  log ${DEBUG} "Issuing service command: '\"${SVC_SCRPT_PTH}\" start \"${user_install_arg}\" \"${APP_DEPLOYMENT_ARGS}\"'"
  "${SVC_SCRPT_PTH}" start "-Duser.install=0" "${APP_DEPLOYMENT_ARGS}" &>/dev/null && success=true

  if [[ "${success}" == true ]]; then
    log ${INFO} "${APP_BASENAME} started successfully"
  else
    log ${WARN} "Failed to start ${SVC_SCRIPT}"
  fi
}

function prompt_to_start_desktop() {
  if [[ "${QUIET}" == false ]] && [[ -n "${DISPLAY}" ]] && [[ "${CLOAKED}" == false ]]; then
    if [[ -n "${SUDO_USER}" ]]; then
      echo -n "Would you like to start ${APP_BASENAME} Desktop as ${SUDO_USER}? (y/n) [y] "
    else
      return 0
    fi

    get_yn_response YN y
    if [[ "${YN}" == "y" ]]; then
      log ${INFO} "Starting desktop UI."
      if [[ -n "${SUDO_USER}" ]]; then
        su "${SUDO_USER}" -c "${GUI_SCRPT_PTH}"
      else
        ${GUI_SCRPT_PTH}
      fi
    fi
  fi
}

function log_install_complete() {
  log ${INFO} ""
  log ${INFO} "Installation is complete. Thank you for installing ${APP_BASENAME}."
  log ${INFO} ""
}

# ===============================================================================
# Begin script
# ===============================================================================
process_input $@
source_install_defaults
set_service_scripts
verify_cmds
check_for_user_install
verify_user
set_install_locations
check_and_handle_prev_install
install_app_prereqs
create_install_dirs
log_install_location_info
create_install_vars
unpack_archive
setup_electron
update_manifest_path
change_log_dir_perms
install_gui_script
install_service_script
install_systemd_script
install_launcher
start_service
log_resource_location
prompt_to_start_desktop
log_install_complete
